/*
    MSound.cpp - Implementation of MSound.dll
    by MUTATE Systems 2017
*/

#pragma comment(lib, "dsound.lib")
#include <windows.h>
#include <mmsystem.h>
#include <string.h>
#include <objbase.h>
#include <stdio.h>
#include <io.h>
#include <math.h>
#include <list>
#include <vector>
#include <dsound.h>
#include "vorbis/codec.h"
#include "vorbis/vorbisfile.h"

#include "MSound.h"

static bool parseWavHeader(const void* pWavFile,
			WAVEFORMATEX** ppwf, BYTE** ppData, DWORD* pcbData) {
	*ppwf = NULL;
	*ppData = NULL;
	*pcbData = 0;

	DWORD* pdwWav  = (DWORD*)pWavFile;
	DWORD dwRiff   = pdwWav[0];
	DWORD dwLength = pdwWav[1];
	DWORD dwType   = pdwWav[2];
    pdwWav += 3;
	if (dwRiff != mmioFOURCC('R', 'I', 'F', 'F') || dwType != mmioFOURCC('W', 'A', 'V', 'E') )
		return false;

	DWORD* pdwEnd = (DWORD*)((BYTE*)pdwWav + dwLength - 4);

	while(pdwWav < pdwEnd) {
		dwType = *pdwWav++;
		dwLength = *pdwWav++;

		switch (dwType) {
		case mmioFOURCC('f', 'm', 't', ' '):
			if ( !*ppwf ) {
				if (dwLength < sizeof(WAVEFORMAT))
					return false;

				*ppwf = (WAVEFORMATEX *)pdwWav;

				if (*ppData  &&  *pcbData)
					return true;
			}
			break;

		case mmioFOURCC('d', 'a', 't', 'a'):
			if ( !*ppData || !*pcbData ) {
				*ppData = (LPBYTE)pdwWav;
				*pcbData = dwLength;
				if (*ppwf)
					return true;
			}
			break;
		}
		pdwWav = (DWORD*)((BYTE*)pdwWav + ((dwLength+1)&~1));
	}
	return false;
}

static bool decodeOgg(const char* szFileName, WAVEFORMATEX* pwf, BYTE** ppData, size_t* lenData) {
	FILE* fpIn = fopen(szFileName, "rb");
	if(!fpIn) return false;

	OggVorbis_File vf;
	if(ov_open(fpIn, &vf, NULL, 0) < 0) {
		fclose(fpIn);
		return false;
	}

	const int DEC_WORD = 16/8; // 16bit
	vorbis_info* vi = ov_info(&vf,-1);
	*lenData = (long)ov_pcm_total(&vf,-1) * vi->channels * DEC_WORD;

	pwf->wFormatTag = WAVE_FORMAT_PCM;
	pwf->nChannels = vi->channels;
	pwf->nSamplesPerSec = vi->rate;
	pwf->nBlockAlign = vi->channels * DEC_WORD;
	pwf->nAvgBytesPerSec = pwf->nBlockAlign * pwf->nSamplesPerSec;
	pwf->wBitsPerSample = DEC_WORD * 8;

    *ppData = new BYTE[*lenData];

	int current_section = 0;
	unsigned nRead = 0;
	int nr = 0;
	while(nRead < *lenData && 
			(nr = ov_read(&vf, (char*)(*ppData) + nRead, *lenData - nRead, 0, DEC_WORD, 1,&current_section)))
		nRead += nr;

	ov_clear(&vf);
	fclose(fpIn);
    return true;
}

static IDirectSound8* g_pDSound;

MSOUND_API int ratioToDecibel(float ratio)
{
    if(ratio >= 1.0f) return DSBVOLUME_MAX; // 0
    if(ratio <= 0) return DSBVOLUME_MIN;
    int r = int(log(ratio) / log(2.0f) * 600);
    return r < DSBVOLUME_MIN ? DSBVOLUME_MIN : r;
}

// Wrapper class of IDirectSoundBuffer
class SoundBuffer : public ISoundPlayer {
    IDirectSoundBuffer* m_pDSB;
public:
    SoundBuffer(void) : m_pDSB(NULL) {}
    bool setupBuffer(WAVEFORMATEX* pwf, const BYTE* pData, size_t len, bool bPlayNonActive);

    void stop(void) override {
        m_pDSB->Stop();
        m_pDSB->SetCurrentPosition(0);
    }
	void stopLoop(void) override {
		DWORD dwStatus = 0;
		if ( m_pDSB->GetStatus(&dwStatus) == DS_OK && (dwStatus & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING )
			m_pDSB->Play(0,0,0);
    }
    void setPan(LONG pan) override {
        m_pDSB->SetPan(pan);
    }
    void setVolume(LONG vol) override {
        m_pDSB->SetVolume(vol);
    }
    void cleanup(void) {
		m_pDSB->Stop();
		m_pDSB->Release();
		m_pDSB = NULL;
    }
    HRESULT copyFrom(SoundBuffer* pOther) {
        return g_pDSound->DuplicateSoundBuffer(pOther->m_pDSB, &m_pDSB);
    }
    bool play(bool bLoop, LONG vol, LONG pan);
};

bool SoundBuffer::play(bool bLoop, LONG vol, LONG pan) {
    if(!m_pDSB) return false;

	DWORD dwStatus = 0;
	if ( m_pDSB->GetStatus(&dwStatus) != DS_OK )
		return false;

	if ( (dwStatus & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING ) {
		m_pDSB->Stop();
		m_pDSB->SetCurrentPosition(0);
	}
	m_pDSB->SetPan(pan);
	m_pDSB->SetVolume(vol);
	m_pDSB->Play(0, 0, bLoop ? DSBPLAY_LOOPING : 0);
    return true;
}

bool SoundBuffer::setupBuffer(WAVEFORMATEX* pwf, const BYTE* pData, size_t len, bool bPlayNonActive) {

    DSBUFFERDESC dsBD = {0};
	dsBD.dwSize = sizeof(dsBD);
	dsBD.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLPAN
		| DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY |
        (bPlayNonActive ? DSBCAPS_GLOBALFOCUS : 0);
	dsBD.dwBufferBytes = len;
	dsBD.lpwfxFormat = pwf;

	if (g_pDSound->CreateSoundBuffer(&dsBD, &m_pDSB, NULL) == DS_OK) {
		LPVOID pMem = NULL;
		DWORD dwSize = 0;
		if ( m_pDSB->Lock(0, dsBD.dwBufferBytes, &pMem, &dwSize, NULL, NULL, 0) == DS_OK ) {
            memcpy(pMem, pData, dwSize);
			m_pDSB->Unlock(pMem, dwSize, NULL, 0);
            return true;
		}
	}

    // Failed
	if(m_pDSB)
		m_pDSB->Release();
    m_pDSB = NULL;
	return false;
}

class SoundEffect : public ISoundEffect {
    std::vector<SoundBuffer> m_buffers;
    std::vector<SoundBuffer>::iterator m_current;
    bool m_bPlayNonActive;

	bool loadWav(const char* FileName);
	bool loadOgg(const char* FileName);
	bool isLoaded(void) { return m_buffers.size() > 0; }
	void cleanup(void);

public:
	SoundEffect(bool bPlayNonActive);
	~SoundEffect();

    // Load sound file (wav or ogg)
	// nClone means the count of the sound that can be play concurrently.
	bool load(const char* FileName, int nClone = 1);

    void release(void);

	// To start the sound.
	// You can control the sound by returned IDirectSoundBuffer.
	ISoundPlayer* play(bool bLoop, LONG vol, LONG pan) override;
};

SoundEffect::SoundEffect(bool bPlayNonActive) :
    m_bPlayNonActive(bPlayNonActive)
{
    m_current = m_buffers.begin();
}

void SoundEffect::cleanup(void) {
	if ( g_pDSound ) {
        for(auto i = m_buffers.begin(); i != m_buffers.end(); i ++)
            (*i).cleanup();
	}
    m_buffers.clear();
}

void SoundEffect::release(void) {
    delete this;
}

SoundEffect::~SoundEffect() {
	cleanup();
}

bool SoundEffect::loadWav(const char* filename) {
	if(!g_pDSound) return 0;

	FILE* fp = fopen(filename, "rb");
	if(!fp)
		return false;
	long len = _filelength(_fileno(fp));

	BYTE* pWavFile = new BYTE[len];
	fread(pWavFile, 1, len, fp);
	fclose(fp);

    WAVEFORMATEX* pwf = NULL;
    DWORD lenData = 0;

	BYTE *pbData = NULL;
	if (!parseWavHeader(pWavFile, &pwf, &pbData, &lenData))
		return false;

    for(auto i = m_buffers.begin(); i != m_buffers.end(); i ++)
        (*i).setupBuffer(pwf, pbData, lenData, m_bPlayNonActive);

    delete[] pWavFile;
	return true;
}

bool SoundEffect::loadOgg(const char* szFileName) {
    BYTE* pmem = 0;
    size_t len = 0;

    WAVEFORMATEX waveFormat = {0};
    if(!decodeOgg(szFileName, &waveFormat, &pmem, &len))
        return false;

    for(auto i = m_buffers.begin(); i != m_buffers.end(); i ++)
        (*i).setupBuffer(&waveFormat, pmem, len, m_bPlayNonActive);

    delete[] pmem;
    return true;
}

bool SoundEffect::load(const char* filename, int nClone) {
    m_buffers.resize(nClone);
    m_current = m_buffers.begin();
	if(loadOgg(filename))
		return true;
    else
    	return loadWav(filename);
}

ISoundPlayer* SoundEffect::play(bool bLoop, LONG vol, LONG pan) {
    if(!m_buffers.size()) return NULL;

    if(m_current == m_buffers.end())
        m_current = m_buffers.begin();

    SoundBuffer* sb = &(*m_current);
    sb->play(bLoop, vol, pan);
    m_current ++;
	return sb;
}

MSOUND_API ISoundEffect* ISoundEffect::create(const char* szFileName, int nClone, bool bPlayNonActive) {
    SoundEffect* pNew = new SoundEffect(bPlayNonActive);
    if(pNew->load(szFileName, nClone))
        return pNew;
    else
        return NULL;
}

class synchronizer {
    CRITICAL_SECTION* m_pcs;
public:
    synchronizer(CRITICAL_SECTION* pcs) {
        m_pcs = pcs;
        EnterCriticalSection(m_pcs);
    }
    ~synchronizer() {
        LeaveCriticalSection(m_pcs);
    }
};

const DWORD Bgm_MinBufferingSec = 4;

class BgmPlayer : public IBgmPlayer {
protected:
    static std::list<BgmPlayer*> m_lInst;
    static CRITICAL_SECTION m_cs;
    static HANDLE m_hThread;
    static DWORD m_dwThreadID;
    static bool m_bTerminate;

    IDirectSoundBuffer* m_pDSB;
    char m_szCurrentFile[_MAX_PATH];
    OggVorbis_File m_OggFile;
    FILE* m_fpOgg;
    WAVEFORMATEX m_WavFmt;
    DWORD m_TotalSizeBytes;
    DWORD m_LoadedBytes;
    bool m_bOggFileReading;
    LONG m_lVolume, m_lPan;
    bool m_bPlayNonActive;

    BgmPlayer(bool bPlayNonActive);
    void loadComplete(void);
    bool loadNext(DWORD posToLoad);
    void onTimer(void);
    static void addInstance(BgmPlayer*);

public:
    static BgmPlayer* create(bool bPlayNonActive);
    bool start(const char* szBgmFileName, bool bLoop) override;
    void stop(void) override;
    void stopLoop(void) override;
    void setVolume(LONG) override;
    void setPan(LONG) override;
    void release(void);

    static void initialize(void);
    static void shutdown(void);
    static void releaseAll(void);
    static DWORD WINAPI procThread(LPVOID param);
};

std::list<BgmPlayer*> BgmPlayer::m_lInst;
CRITICAL_SECTION BgmPlayer::m_cs;
DWORD BgmPlayer::m_dwThreadID;
bool BgmPlayer::m_bTerminate;
HANDLE BgmPlayer::m_hThread;

BgmPlayer::BgmPlayer(bool bPlayNonActive) :
    m_bPlayNonActive(bPlayNonActive),
    m_pDSB(NULL),
    m_fpOgg(NULL),
    m_TotalSizeBytes(0),
    m_LoadedBytes(0),
    m_bOggFileReading(false),
    m_lVolume(DSBVOLUME_MAX),
    m_lPan(0)
{
    m_szCurrentFile[0] = 0;
}

void BgmPlayer::initialize(void) {
	InitializeCriticalSection(&m_cs);
    m_hThread = CreateThread(NULL , 0 , procThread , NULL , 0 , &m_dwThreadID);
}

void BgmPlayer::shutdown(void) {
    BgmPlayer::releaseAll();
	m_bTerminate = true;
    WaitForSingleObject(m_hThread, INFINITE);
	DeleteCriticalSection(&m_cs);
}

void BgmPlayer::loadComplete(void) {
	if(m_bOggFileReading) {
		m_bOggFileReading = false;
		ov_clear(&m_OggFile);
	}
	if(m_fpOgg) {
		fclose(m_fpOgg);
		m_fpOgg = NULL;
	}
}

bool BgmPlayer::loadNext(DWORD posToLoad) {
	if(m_bOggFileReading) {
		bool bFinal = false;
		if(posToLoad >= m_TotalSizeBytes) {
			posToLoad = m_TotalSizeBytes;
			bFinal = true;
		}
		int sizeLock = (int)posToLoad - m_LoadedBytes;
		
		if(sizeLock > 0) {
			LPVOID pMem = NULL;
			DWORD dwSize = 0;
			if ( m_pDSB->Lock(m_LoadedBytes, sizeLock, &pMem, &dwSize, NULL, NULL, 0) == DS_OK ) {
				int current_section = 0;
				unsigned nRead = 0;
				int nr = 0;
				while(nRead < dwSize && 
						(nr = ov_read(&m_OggFile, (char*)pMem + nRead, dwSize - nRead, 0, 16/8, 1, &current_section))) {
					nRead += nr;
				}
				m_pDSB->Unlock(pMem, dwSize, NULL, 0);
			}
			m_LoadedBytes = posToLoad;
		}
		if(bFinal) {
			loadComplete();
		}
	}
	return true;
}

bool BgmPlayer::start(const char* szBgmFileName, bool bLoop) {

    if(!g_pDSound) return false;

    if(m_pDSB && !stricmp(m_szCurrentFile, szBgmFileName)) return true;
    
    synchronizer sy(&m_cs);
	strcpy(m_szCurrentFile, szBgmFileName);

	stop();

	if(m_bOggFileReading) {
		ov_clear(&m_OggFile);
		fclose(m_fpOgg);
	}
	memset(&m_OggFile, 0, sizeof(m_OggFile));

	m_fpOgg = fopen(szBgmFileName, "rb");
	if(!m_fpOgg) return false;

	if(ov_open(m_fpOgg, &m_OggFile, NULL, 0) < 0) {
		fclose(m_fpOgg);
		return false;
	}

	const int DEC_WORD = 16/8; // 16bit
	vorbis_info* vi = ov_info(&m_OggFile, -1);
	m_TotalSizeBytes = (long)ov_pcm_total(&m_OggFile, -1) * vi->channels * DEC_WORD;

	ZeroMemory(&m_WavFmt, sizeof(m_WavFmt));
	m_WavFmt.wFormatTag = WAVE_FORMAT_PCM;
	m_WavFmt.nChannels = vi->channels;
	m_WavFmt.nSamplesPerSec = vi->rate;
	m_WavFmt.nBlockAlign = vi->channels * DEC_WORD;
	m_WavFmt.nAvgBytesPerSec = m_WavFmt.nBlockAlign * m_WavFmt.nSamplesPerSec;
	m_WavFmt.wBitsPerSample = DEC_WORD * 8;

	DSBUFFERDESC dsBD;
	ZeroMemory( &dsBD, sizeof dsBD );
	dsBD.dwSize = sizeof(dsBD);
	dsBD.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLPAN
		| DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | 
        (m_bPlayNonActive ? DSBCAPS_GLOBALFOCUS : 0);
	dsBD.dwBufferBytes = m_TotalSizeBytes;
	dsBD.lpwfxFormat = &m_WavFmt;

	m_pDSB = NULL;
	bool bSuccess = false;
	if (g_pDSound->CreateSoundBuffer(&dsBD, &m_pDSB, NULL) == DS_OK) {
		m_bOggFileReading = true;
		m_LoadedBytes = 0;
		loadNext(m_WavFmt.nAvgBytesPerSec * Bgm_MinBufferingSec);
		setVolume(m_lVolume);
		m_pDSB->Play(0, 0, bLoop ? DSBPLAY_LOOPING : 0);
	}
	else {
		loadComplete();
		return false;
	}
	return true;
}

void BgmPlayer::stop(void) {
    if(!g_pDSound) return;
    synchronizer sy(&m_cs);
	loadComplete();
	if(m_pDSB) {
		m_pDSB->Stop();
		m_pDSB->Release();
		m_pDSB = NULL;
	}
	*m_szCurrentFile = 0;
}

void BgmPlayer::stopLoop(void) {
    DWORD dwStatus = 0;
	if ( m_pDSB && m_pDSB->GetStatus(&dwStatus) == DS_OK && (dwStatus & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING )
		m_pDSB->Play(0,0,0);
}

void BgmPlayer::setVolume(LONG vol) {
    m_lVolume = vol;
	if(m_pDSB) m_pDSB->SetVolume(vol);
}

void BgmPlayer::setPan(LONG pan) {
    m_lPan = pan;
	if(m_pDSB) m_pDSB->SetPan(pan);
}

void BgmPlayer::onTimer(void) {
	if(m_bOggFileReading) {
		DWORD dwCurrent = 0, dwWrite = 0;
		m_pDSB->GetCurrentPosition(&dwCurrent, &dwWrite);
		loadNext(dwCurrent + m_WavFmt.nAvgBytesPerSec * Bgm_MinBufferingSec);
	}
}

DWORD WINAPI BgmPlayer::procThread(LPVOID param) {
	while(!m_bTerminate) {
        {
            synchronizer sy(&m_cs);
            for(auto i = m_lInst.begin(); i != m_lInst.end(); i ++)
                (*i)->onTimer();
        }
		Sleep(500);
	}
	return 0;
}

void BgmPlayer::addInstance(BgmPlayer* p) {
    synchronizer sy(&m_cs);
    m_lInst.push_back(p);
}

BgmPlayer* BgmPlayer::create(bool bPlayNonActive) {
    BgmPlayer* newInstance = new BgmPlayer(bPlayNonActive);
    addInstance(newInstance); 
    return newInstance;
}

void BgmPlayer::release(void) {
    stop();
    {
        synchronizer sy(&m_cs);
        m_lInst.remove(this);
    }
    delete this;
}

void BgmPlayer::releaseAll(void) {
    {
        synchronizer sy(&m_cs);
        for(auto i = m_lInst.begin(); i != m_lInst.end(); i ++) {
            (*i)->stop();
            delete (*i);
        }
    }
    m_lInst.clear();
}

MSOUND_API IBgmPlayer* IBgmPlayer::create(bool bPlayNonActive) {
    return BgmPlayer::create(bPlayNonActive);
}

MSOUND_API bool MSoundInitialize(HWND hwnd) {
	if ( g_pDSound ) return true;

	if ( DirectSoundCreate8(NULL, &g_pDSound, NULL) != DS_OK ) {
		g_pDSound = NULL;
		return false;
	}

	g_pDSound->SetCooperativeLevel(hwnd, DSSCL_NORMAL);
	DSBUFFERDESC dsbd = {0};
	dsbd.dwSize = sizeof dsbd;
	dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;
	IDirectSoundBuffer* pPrimary = NULL;
	g_pDSound->CreateSoundBuffer(&dsbd, &pPrimary, NULL );

    BgmPlayer::initialize();
    return true;
}

MSOUND_API void MSoundShutdown(void) {
    BgmPlayer::shutdown();

	if ( g_pDSound ) {
		g_pDSound->Release();
		g_pDSound = 0;
	}
}

BOOL APIENTRY DllMain( HMODULE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
					 )
{
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
	return TRUE;
}
